-- Format:
-- * = regular wildcard, 0 - infinite
-- ? = single character wildcard
-- ????* = 4 character + (not hardcoded, logical consequence of the above 2)
-- *? = 0 - 1 characters
-- ** = '*'
-- *?*?*??? = 2 - 5 characters (not coded, just logic again)

-- The actual characters above are defaults, not locked in
-- Patterns are interpreted left to right

include rmisc.e

without warning

global integer WildcardChar, SingleWildChar

    WildcardChar = '*'
    SingleWildChar = '?'


constant wildcard = -1, infinite = 0, spc_char = -2

sequence sections
integer browse, len

function int_wild(sequence pat)
    sequence sec
    
    if pat[browse] = WildcardChar then
        browse += 1
        if browse > len then
            sections = append(sections, { wildcard, 0, infinite })
            return 1
        elsif pat[browse] = WildcardChar then
            browse -= 1
            return 0
        elsif pat[browse] = SingleWildChar then
            sec = { wildcard, 0, 1 }
        else
            sections = append(sections, { wildcard, 0, infinite })
            return 1
        end if
    elsif pat[browse] = SingleWildChar then
        sec = { wildcard, 1, 1 }
    else
        return 0
    end if
    
    browse += 1
    
    while browse <= len do
        if pat[browse] = WildcardChar then
            browse += 1
            if browse > len then
                sec[3] = infinite
                exit
            elsif pat[browse] = WildcardChar then
                browse -= 1
                exit
            elsif pat[browse] = SingleWildChar then
                sec[3] += 1
            else
                sec[3] = infinite
                exit
            end if
        elsif pat[browse] = SingleWildChar then
            sec[2..3] += 1
        else
            exit
        end if
        
        browse += 1
    end while
    
    sections = append(sections, sec)
    return 1
end function

function int_text(sequence pat)
    sequence sec
    
    sec = ""
    
    while browse <= len and pat[browse] != SingleWildChar do

        if pat[browse] = WildcardChar then
            if browse < len and pat[browse + 1] = WildcardChar then
                browse += 1
            else
                exit
            end if
        end if
        sec = append(sec, pat[browse])
        browse += 1
    end while
    
    if length(sec) then
        sections = append(sections, sec)
        return 1
    end if
    
    return 0
end function

global function InterpretPattern(sequence pattern)

    len = length(pattern)

    sections = {}
    browse = 1
    
    while browse <= len and (int_wild(pattern) or int_text(pattern)) do end while
    
    return sections

end function


global function WildcardMatch_real(sequence pat, sequence str)
    sequence pos, ret
    integer lp, ls, j, lpj, lm

    if pat[1][1] != wildcard and match(pat[1], str) != 1 then
        return 0
    end if

    -- Endpoint control, explained later    
    str = append(str, spc_char)
    pat = append(pat, { spc_char } )


    lp = length(pat)
    ls = length(str)
    pos = repeat({}, lp)
    
    pos[1] = { 1 }
    
    for i = 2 to lp do
        j = i - 1
        lpj = length(pos[j])

        -- calculation of potential possibilities
        if pat[j][1] = wildcard then
            if pat[j][3] = infinite then
                pos[i] = pos[j][1] + pat[j][2]
                pos[i] = repeat(pos[i], ls - pos[i] + 1)
                for p = 2 to length(pos[i]) do
                    pos[i][p] += p - 1
                end for
            else
                lm = pat[j][2] -- length minimum
                pos[i] = repeat(0, lpj * (1 + pat[j][3] - lm))
                
                for p = lm to pat[j][3] do
                    pos[i][1 + (p - lm) * lpj..(1 + p - lm) * lpj] = pos[j] + p
                end for
            end if
        else
            pos[i] = pos[j] + length(pat[j])
        end if
        
        -- sorting by certain limits, being current minimum position and
        -- string length
        
        -- Do we know that entry 1 in pos[i] is the lowest value?
        -- wildcard infinite: yes, explicitly assigned greater values later
        -- wildcard defined : given that it was so alreay, yes.
        --                      p is an increasing value, pos[j] starts low
        -- literal string : given the same condition
        -- AND THAT CONDITION IS GIVEN IN pos[1] = { 1 }
        
        -- so we have smallest value first
                
        if pat[i][1] = wildcard then
            pos[i] = Arrange(pos[i], pos[i][1], 1 + ls - pat[i][2])
            -- arrange pos[i] low to high, only 1 of each value, only values
            -- within range of the string, and our current minimum location
            -- pat[i][2] = minimum length
            -- minimum length 0 works, 1 + ls becomes the limit
        else
            pos[i] = Arrange(pos[i], pos[i][1], 1 + ls - length(pat[i]))
            -- arrange pos[i] low to high, only 1 of each value, only values
            -- within range of the string, and our current minimum location

            pos[i] = Shared(pos[i], MatchList(pat[i], str, pos[i][1]))
            -- checks what positions that match both in the string, and in the
            -- position list. Cuts out unavailable positions.
        end if
        
        if not length(pos[i]) then return 0 end if
    end for
    
    -- Now, we've built a list over possible positions, and several combinations
    -- might all work out. We're going to find one, and one only, by starting at
    -- the right end (high index), where the possible index list has been built
    -- on data influenced by all the other pattern sections.
    
    -- That is a limiting influence.
    -- Any values listed in one entry is compatible with that to the left,
    -- and this sequentially means that every entry in the last field is
    -- compatible everywhere.
    
    -- What remains to be seen is if it is compatible with the endpoint of the
    -- string. If it is long enough, in any of the valid positions.
    
    -- The affinity goes towards assigning every segment of the pattern as low
    -- and index as possible in the string.

    -- ENDPOINT CONTROL
    
    -- Because the character "spc_char" only exists in one location in the
    -- string, it will fail to be assigned, unless the exact endpoint can be
    -- reached. If so, we will not have reached this location. We know it'll
    -- work at this point.  
    
    -- Now, let us get those positions finalized
    
    pos[lp] = pos[lp][1]
    ret = repeat("", lp - 1)
    
    for i = lp - 1 to 1 by -1 do
        j = i + 1
        if pat[i][1] = wildcard then
            if pat[i][3] = infinite then
                pos[i] = pos[i][1]
            else
                for p = pos[j] - pat[i][3] to pos[j] - pat[i][2]  do
                    if find(p, pos[i]) then
                        pos[i] = p  -- it is assumed that this will always
                        exit        -- occur, not fully analyzed yet
                    end if
                end for
            end if
        else
            pos[i] = pos[j] - length(pat[i])
        end if
        
        ret[i] = str[pos[i]..pos[j] - 1]
    end for 
    
    return ret
end function

global function WildcardMatch(sequence pattern, sequence string)
    return WildcardMatch_real(InterpretPattern(pattern), string)
end function
